




class OverloadedBalanceOfContractFunction0(PypechainContractFunction):
    """ContractFunction for the balanceOf() method."""

    _function_name = "balanceOf"
    _type_signature = expand_struct_type_str(
        tuple([]),
        structs
    )

    
    
    

    

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier | None = None,
        state_override: StateOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> int:
        """returns int."""
        # We handle the block identifier = None case here for typing.
        if block_identifier is None:
            block_identifier = self.w3.eth.default_block

        
        # Define the expected return types from the smart contract call
        return_types = int
        

        # Call the function
        try:
            raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="call",
                transaction=transaction,
                block_identifier=block_identifier,
            ) from err
        

        
        return cast(int, rename_returned_types(structs, return_types, raw_values))
        

    def transact(self, transaction: TxParams | None = None) -> HexBytes:
        try:
            return super().transact(transaction)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="transact",
                transaction=transaction,
                block_identifier="pending", # race condition here, best effort to get block of txn.
            ) from err

    def estimate_gas(
        self, 
        transaction: TxParams | None = None, 
        block_identifier: BlockIdentifier | None = None, 
        state_override: StateOverride | None = None,
    ) -> int:
        try:
            return super().estimate_gas(transaction, block_identifier, state_override)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="build",
                transaction=transaction,
                block_identifier="pending", # race condition here, best effort to get block of txn.
            ) from err
    
    def build_transaction(self, transaction: TxParams | None = None) -> TxParams:
        try:
            return super().build_transaction(transaction)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="build",
                transaction=transaction,
                block_identifier="pending", # race condition here, best effort to get block of txn.
            ) from err
    
    def sign_and_transact(self, account: LocalAccount, transaction: TxParams | None = None) -> HexBytes:
        """Convenience method for signing and sending a transaction using the provided account.

        Arguments
        ---------
        account : LocalAccount
            The account to use for signing and sending the transaction.
        transaction : TxParams | None, optional
            The transaction parameters to use for sending the transaction.
        
        Returns
        -------
        HexBytes
            The transaction hash.
        """
        if transaction is None:
            transaction_params: TxParams = {}
        else:
            transaction_params: TxParams = transaction

        if "from" in transaction_params:
            # Ensure if transaction is set, it matches
            assert transaction_params["from"] == account.address, f"Transaction from {transaction_params['from']} does not match account {account.address}"
        else:
            transaction_params["from"] = account.address

        if "gas" not in transaction_params:
            # Web3 default gas estimate seems to be underestimating gas, likely due to
            # not looking at pending block. Here, we explicitly call estimate gas
            # if gas isn't passed in.
            transaction_params["gas"] = self.estimate_gas(transaction_params, block_identifier="pending")
        
        # Build the raw transaction
        raw_transaction = self.build_transaction(transaction_params)

        if "nonce" not in raw_transaction:
            raw_transaction["nonce"] = self.w3.eth.get_transaction_count(account.address, block_identifier="pending")

        # Sign the raw transaction
        # Mismatched types between account and web3py
        signed_transaction = account.sign_transaction(raw_transaction) # type: ignore

        # Send the signed transaction
        try:
            return self.w3.eth.send_raw_transaction(signed_transaction.raw_transaction)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="transact",
                transaction=transaction_params,
                block_identifier="pending", # race condition here, best effort to get block of txn.
            ) from err
        
    def sign_transact_and_wait(
        self, 
        account: LocalAccount, 
        transaction: TxParams | None = None, 
        timeout: float | None = None, 
        poll_latency: float | None = None, 
        validate_transaction: bool = False,
    ) -> TxReceipt:
        """Convenience method for signing and sending a transaction using the provided account.

        Arguments
        ---------
        account : LocalAccount
            The account to use for signing and sending the transaction.
        transaction : TxParams | None, optional
            The transaction parameters to use for sending the transaction.
        timeout: float, optional
            The number of seconds to wait for the transaction to be mined. Defaults to 120.
        poll_latency: float, optional
            The number of seconds to wait between polling for the transaction receipt. Defaults to 0.1.
        validate_transaction: bool, optional
            Whether to validate the transaction. If True, will throw an exception if the resulting
            tx_receipt returned a failure status.

        Returns
        -------
        HexBytes
            The transaction hash.
        """

        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-positional-arguments

        if timeout is None:
            timeout = 120
        if poll_latency is None:
            poll_latency = 0.1

        tx_hash = self.sign_and_transact(account, transaction)
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash, timeout=timeout, poll_latency=poll_latency)
        # Check the receipt, throwing an error if status == 0
        if validate_transaction:
            return check_txn_receipt(self, tx_hash, tx_receipt)
        else:
            return tx_receipt

        


class OverloadedBalanceOfContractFunction1(PypechainContractFunction):
    """ContractFunction for the balanceOf(str) method."""

    _function_name = "balanceOf"
    _type_signature = expand_struct_type_str(
        tuple(['str']),
        structs
    )

    
    
    

    

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier | None = None,
        state_override: StateOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> int:
        """returns int."""
        # We handle the block identifier = None case here for typing.
        if block_identifier is None:
            block_identifier = self.w3.eth.default_block

        
        # Define the expected return types from the smart contract call
        return_types = int
        

        # Call the function
        try:
            raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="call",
                transaction=transaction,
                block_identifier=block_identifier,
            ) from err
        

        
        return cast(int, rename_returned_types(structs, return_types, raw_values))
        

    def transact(self, transaction: TxParams | None = None) -> HexBytes:
        try:
            return super().transact(transaction)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="transact",
                transaction=transaction,
                block_identifier="pending", # race condition here, best effort to get block of txn.
            ) from err

    def estimate_gas(
        self, 
        transaction: TxParams | None = None, 
        block_identifier: BlockIdentifier | None = None, 
        state_override: StateOverride | None = None,
    ) -> int:
        try:
            return super().estimate_gas(transaction, block_identifier, state_override)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="build",
                transaction=transaction,
                block_identifier="pending", # race condition here, best effort to get block of txn.
            ) from err
    
    def build_transaction(self, transaction: TxParams | None = None) -> TxParams:
        try:
            return super().build_transaction(transaction)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="build",
                transaction=transaction,
                block_identifier="pending", # race condition here, best effort to get block of txn.
            ) from err
    
    def sign_and_transact(self, account: LocalAccount, transaction: TxParams | None = None) -> HexBytes:
        """Convenience method for signing and sending a transaction using the provided account.

        Arguments
        ---------
        account : LocalAccount
            The account to use for signing and sending the transaction.
        transaction : TxParams | None, optional
            The transaction parameters to use for sending the transaction.
        
        Returns
        -------
        HexBytes
            The transaction hash.
        """
        if transaction is None:
            transaction_params: TxParams = {}
        else:
            transaction_params: TxParams = transaction

        if "from" in transaction_params:
            # Ensure if transaction is set, it matches
            assert transaction_params["from"] == account.address, f"Transaction from {transaction_params['from']} does not match account {account.address}"
        else:
            transaction_params["from"] = account.address

        if "gas" not in transaction_params:
            # Web3 default gas estimate seems to be underestimating gas, likely due to
            # not looking at pending block. Here, we explicitly call estimate gas
            # if gas isn't passed in.
            transaction_params["gas"] = self.estimate_gas(transaction_params, block_identifier="pending")
        
        # Build the raw transaction
        raw_transaction = self.build_transaction(transaction_params)

        if "nonce" not in raw_transaction:
            raw_transaction["nonce"] = self.w3.eth.get_transaction_count(account.address, block_identifier="pending")

        # Sign the raw transaction
        # Mismatched types between account and web3py
        signed_transaction = account.sign_transaction(raw_transaction) # type: ignore

        # Send the signed transaction
        try:
            return self.w3.eth.send_raw_transaction(signed_transaction.raw_transaction)
        except Exception as err: # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=OverloadedContractErrors,
                err=err,
                contract_call_type="transact",
                transaction=transaction_params,
                block_identifier="pending", # race condition here, best effort to get block of txn.
            ) from err
        
    def sign_transact_and_wait(
        self, 
        account: LocalAccount, 
        transaction: TxParams | None = None, 
        timeout: float | None = None, 
        poll_latency: float | None = None, 
        validate_transaction: bool = False,
    ) -> TxReceipt:
        """Convenience method for signing and sending a transaction using the provided account.

        Arguments
        ---------
        account : LocalAccount
            The account to use for signing and sending the transaction.
        transaction : TxParams | None, optional
            The transaction parameters to use for sending the transaction.
        timeout: float, optional
            The number of seconds to wait for the transaction to be mined. Defaults to 120.
        poll_latency: float, optional
            The number of seconds to wait between polling for the transaction receipt. Defaults to 0.1.
        validate_transaction: bool, optional
            Whether to validate the transaction. If True, will throw an exception if the resulting
            tx_receipt returned a failure status.

        Returns
        -------
        HexBytes
            The transaction hash.
        """

        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-positional-arguments

        if timeout is None:
            timeout = 120
        if poll_latency is None:
            poll_latency = 0.1

        tx_hash = self.sign_and_transact(account, transaction)
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash, timeout=timeout, poll_latency=poll_latency)
        # Check the receipt, throwing an error if status == 0
        if validate_transaction:
            return check_txn_receipt(self, tx_hash, tx_receipt)
        else:
            return tx_receipt

        

class OverloadedBalanceOfContractFunction(PypechainContractFunction):
    """ContractFunction for the balanceOf method."""
    # super() call methods are generic, while our version adds values & types
    # pylint: disable=arguments-differ# disable this warning when there is overloading
    # pylint: disable=function-redefined

    _function_name = "balanceOf"


    # Make lookup for function signature -> overloaded function
    # The function signatures are python types, as we need to do a
    # lookup of arguments passed in to contract function
    _functions: dict[str, PypechainContractFunction]



    @overload
    def __call__(self) -> OverloadedBalanceOfContractFunction0:  # type: ignore
        ...

    @overload
    def __call__(self, who: str) -> OverloadedBalanceOfContractFunction1:  # type: ignore
        ...


    def __call__(self, *args, **kwargs) -> OverloadedBalanceOfContractFunction:  # type: ignore
        clone = super().__call__(*(dataclass_to_tuple(arg) for arg in args), **{key: dataclass_to_tuple(arg) for key, arg in kwargs.items()})

        # Arguments is the flattened set of arguments from args and kwargs, ordered by the abi
        # We get the python types of the args passed in, but remapped from tuples -> dataclasses
        arg_types = get_arg_type_names(clone.arguments)

        # Look up the function class based on arg types.
        # We ensure we use a copy of the original object.
        function_obj = copy.copy(self._functions[arg_types])

        function_obj.args = clone.args
        function_obj.kwargs = clone.kwargs

        # The `@overload` of `__call__` takes care of setting the type of this object correctly
        return function_obj # type: ignore
    
    @classmethod
    def factory(
        cls, class_name: str, **kwargs: Any
    ) -> Self:
        out = super().factory(class_name, **kwargs)

        # We initialize our overridden functions here.
        # Note that we use the initialized object to ensure each function
        # is attached to the instanciated object 
        # (attached to a specific web3 and contract address)
        out._functions = {
        
            OverloadedBalanceOfContractFunction0._type_signature: OverloadedBalanceOfContractFunction0.factory(
                "OverloadedBalanceOfContractFunction0",
                **kwargs
            ),
        
            OverloadedBalanceOfContractFunction1._type_signature: OverloadedBalanceOfContractFunction1.factory(
                "OverloadedBalanceOfContractFunction1",
                **kwargs
            ),
        
        }
        return out





class OverloadedContractFunctions(ContractFunctions):
    """ContractFunctions for the Overloaded contract."""

    balanceOf: OverloadedBalanceOfContractFunction

    def __init__(
         self,
         abi: ABI,
         w3: "Web3",
         address: ChecksumAddress | None = None,
         decode_tuples: bool | None = False,
     ) -> None:
         super().__init__(abi, w3, address, decode_tuples)
         self.balanceOf = OverloadedBalanceOfContractFunction.factory(
             "balanceOf",
             w3=w3,
             contract_abi=abi,
             address=address,
             decode_tuples=decode_tuples,
             abi_element_identifier="balanceOf",
         )
         


"""Utilities for working with foundry-rs."""

from typing import TypeGuard

from web3 import Web3

from pypechain.foundry.types import FoundryJson
from pypechain.utilities.types import LinkReferences


def is_foundry_json(val: object) -> TypeGuard[FoundryJson]:
    """Determines whether a json object is a FoundryJson."""
    required_keys = {"abi", "bytecode", "deployedBytecode", "methodIdentifiers", "rawMetadata", "metadata", "id"}
    return isinstance(val, dict) and required_keys.issubset(val.keys())


def get_bytecode_from_foundry_json(json_abi: FoundryJson) -> str:
    """Gets the bytecode from a foundry json file."""
    return json_abi.get("bytecode").get("object")


def get_bytecode_link_references_from_foundry_json(json_abi: FoundryJson) -> list[LinkReferences]:
    """Gets the bytecode from a foundry json file."""
    link_references = json_abi.get("bytecode").get("linkReferences")

    # Rework link references to be a dictionary keyed by a contract name, valued with
    # the hex code placeholder in the bytecode.
    out_link_references = []
    for function_path, value in link_references.items():
        for contract_name, _ in value.items():
            fully_qualified_lib_name = f"{function_path}:{contract_name}"
            hex_code = Web3.keccak(text=fully_qualified_lib_name).hex()
            # Hex returns raw hex without 0x. We take the 34 character prefix, and add the placeholder
            # token
            placehoder_code = f"__${hex_code[0:34]}$__"
            # TODO there may be scoping issues here
            out_link_references.append(LinkReferences(contract_name=contract_name, placeholder_code=placehoder_code))

    return out_link_references



"""Dataclasses for all structs in the StructsC contract.

DO NOT EDIT.  This file was generated by pypechain v0.0.46.
See documentation at https://github.com/delvtech/pypechain """

# super() call methods are generic, while our version adds values & types
# pylint: disable=arguments-differ

# contracts have PascalCase names
# pylint: disable=invalid-name

# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments

# unable to determine which imports will be used in the generated code
# pylint: disable=unused-import

# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return

# We import this contract itself to ensure all nested structs have a fully qualified name.
# We use this to avoid namespace collisions, as well as having a uniform
# type structure to do lookups when functions return these structs.
# pylint: disable=import-self


from __future__ import annotations

from dataclasses import dataclass

from ..IStructs import IStructsTypes as IStructs
from ..StructsC import StructsCTypes as StructsC


@dataclass
class InnyStruct:
    """InnyStruct struct."""

    innerVal: int


@dataclass
class NestedStruct:
    """NestedStruct struct."""

    inner: StructsC.InnyStruct


@dataclass
class OuterStruct:
    """OuterStruct struct."""

    nested: StructsC.NestedStruct


@dataclass
class CStruct:
    """CStruct struct."""

    inner: IStructs.InnerStruct





class OverloadedTransferContractEvent(ContractEvent):
    """ContractEvent for Transfer."""
    # super() get_logs and create_filter methods are generic, while our version adds values & types
    # pylint: disable=arguments-differ

    
    
    # pylint: disable=useless-parent-delegation
    def __init__(self, *argument_names: tuple[str]) -> None:
        super().__init__(*argument_names)

    @combomethod_typed
    def get_logs_typed(
        self,
        argument_filters: dict[str, Any] | None = None,
        from_block: BlockIdentifier | None = None,
        to_block: BlockIdentifier | None = None,
        block_hash: HexBytes | None = None,
    ) -> Iterable[TransferEvent]:
        """Extension of `get_logs` that return a typed dataclass of the event."""
        abi_events = super().get_logs(argument_filters=argument_filters, from_block=from_block, to_block=to_block, block_hash=block_hash)
        # TODO there may be issues with this function if the user uses a middleware that changes event structure.
        return [
            TransferEvent(
                log_index = abi_event.logIndex,
                transaction_index = abi_event.transactionIndex,
                transaction_hash = abi_event.transactionHash,
                address = abi_event.address,
                block_hash = abi_event.blockHash,
                block_number = abi_event.blockNumber,

                args = TransferEvent.TransferEventArgs(
                    _from = abi_event.args["_from"],
                    to = abi_event.args["to"],
                    value = abi_event.args["value"],
                ),

            ) for abi_event in abi_events
        ]
    
    @combomethod_typed
    def process_receipt_typed(
        self, txn_receipt: TxReceipt, errors: EventLogErrorFlags = WARN
    ) -> Iterable[TransferEvent]:
        """Extension of `process_receipt` that return a typed dataclass of the event."""
        abi_events = super().process_receipt(txn_receipt, errors)
        # TODO there may be issues with this function if the user uses a middleware that changes event structure.
        return [
            TransferEvent(
                log_index = abi_event.logIndex,
                transaction_index = abi_event.transactionIndex,
                transaction_hash = abi_event.transactionHash,
                address = abi_event.address,
                block_hash = abi_event.blockHash,
                block_number = abi_event.blockNumber,

                args = TransferEvent.TransferEventArgs(
                    _from = abi_event.args["_from"],
                    to = abi_event.args["to"],
                    value = abi_event.args["value"],
                ),

            ) for abi_event in abi_events
        ]

    @combomethod_typed
    def create_filter( # type: ignore
        self,
        *,  # PEP 3102
        argument_filters: dict[str, Any] | None = None,
        from_block: BlockIdentifier | None = None,
        to_block: BlockIdentifier = "latest",
        address: ChecksumAddress | None = None,
        topics: Sequence[Any] | None = None,
    ) -> LogFilter:
        return cast(LogFilter, super().create_filter(argument_filters=argument_filters, from_block=from_block, to_block=to_block, address=address, topics=topics))



class OverloadedApprovalContractEvent(ContractEvent):
    """ContractEvent for Approval."""
    # super() get_logs and create_filter methods are generic, while our version adds values & types
    # pylint: disable=arguments-differ

    
    
    # pylint: disable=useless-parent-delegation
    def __init__(self, *argument_names: tuple[str]) -> None:
        super().__init__(*argument_names)

    @combomethod_typed
    def get_logs_typed(
        self,
        argument_filters: dict[str, Any] | None = None,
        from_block: BlockIdentifier | None = None,
        to_block: BlockIdentifier | None = None,
        block_hash: HexBytes | None = None,
    ) -> Iterable[ApprovalEvent]:
        """Extension of `get_logs` that return a typed dataclass of the event."""
        abi_events = super().get_logs(argument_filters=argument_filters, from_block=from_block, to_block=to_block, block_hash=block_hash)
        # TODO there may be issues with this function if the user uses a middleware that changes event structure.
        return [
            ApprovalEvent(
                log_index = abi_event.logIndex,
                transaction_index = abi_event.transactionIndex,
                transaction_hash = abi_event.transactionHash,
                address = abi_event.address,
                block_hash = abi_event.blockHash,
                block_number = abi_event.blockNumber,

                args = ApprovalEvent.ApprovalEventArgs(
                    owner = abi_event.args["owner"],
                    spender = abi_event.args["spender"],
                    value = abi_event.args["value"],
                ),

            ) for abi_event in abi_events
        ]
    
    @combomethod_typed
    def process_receipt_typed(
        self, txn_receipt: TxReceipt, errors: EventLogErrorFlags = WARN
    ) -> Iterable[ApprovalEvent]:
        """Extension of `process_receipt` that return a typed dataclass of the event."""
        abi_events = super().process_receipt(txn_receipt, errors)
        # TODO there may be issues with this function if the user uses a middleware that changes event structure.
        return [
            ApprovalEvent(
                log_index = abi_event.logIndex,
                transaction_index = abi_event.transactionIndex,
                transaction_hash = abi_event.transactionHash,
                address = abi_event.address,
                block_hash = abi_event.blockHash,
                block_number = abi_event.blockNumber,

                args = ApprovalEvent.ApprovalEventArgs(
                    owner = abi_event.args["owner"],
                    spender = abi_event.args["spender"],
                    value = abi_event.args["value"],
                ),

            ) for abi_event in abi_events
        ]

    @combomethod_typed
    def create_filter( # type: ignore
        self,
        *,  # PEP 3102
        argument_filters: dict[str, Any] | None = None,
        from_block: BlockIdentifier | None = None,
        to_block: BlockIdentifier = "latest",
        address: ChecksumAddress | None = None,
        topics: Sequence[Any] | None = None,
    ) -> LogFilter:
        return cast(LogFilter, super().create_filter(argument_filters=argument_filters, from_block=from_block, to_block=to_block, address=address, topics=topics))



class OverloadedContractEvents(ContractEvents):
    """ContractEvents for the Overloaded contract."""

    Transfer: Type[OverloadedTransferContractEvent]

    Approval: Type[OverloadedApprovalContractEvent]


    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: ChecksumAddress | None = None,
    ) -> None:
        super().__init__(abi, w3, address)
        self.Transfer = cast(Type[OverloadedTransferContractEvent], OverloadedTransferContractEvent.factory(
            "Transfer",
            w3=w3,
            contract_abi=abi,
            address=address,
            event_name="Transfer"
        ))
        self.Approval = cast(Type[OverloadedApprovalContractEvent], OverloadedApprovalContractEvent.factory(
            "Approval",
            w3=w3,
            contract_abi=abi,
            address=address,
            event_name="Approval"
        ))
        


"""Example showcasing pypechain."""

from __future__ import annotations

import os
from pathlib import Path

import pytest
from web3 import Web3
from web3.contract.contract import Contract

from example.types.Example import ExampleContract, InnerStruct, NestedStruct, SimpleStruct
from pypechain.utilities.abi import load_abi_infos_from_file

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name
# pylint: disable=invalid-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


@pytest.mark.usefixtures("process_contracts")
class TestExampleContract:
    """Tests pipeline from bots making trades to viewing the trades in the db"""

    def test_using_vanilla_web3py(self, w3: Web3):
        """Tests with vanilla web3py"""

        # Get the directory of the current script
        script_dir = os.path.dirname(os.path.abspath(__file__))
        # Construct the path to the file relative to the script directory
        file_path = Path(os.path.join(script_dir, "abis", "Example.sol/Example.json"))
        # Open the json and grab the abi and bytecode
        abi_info = load_abi_infos_from_file(file_path)[0]

        ExampleContract = w3.eth.contract(abi=abi_info.abi, bytecode=abi_info.bytecode)

        # Submit the transaction that deploys the contract
        tx_hash = ExampleContract.constructor("example").transact()
        # Wait for the transaction to be mined, and get the transaction receipt
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        deployed_contract: Contract = w3.eth.contract(
            address=tx_receipt.contractAddress, abi=abi_info.abi  # type: ignore
        )

        result = deployed_contract.functions.flipFlop(1, 2).call()
        assert result == [2, 1]

        crazy_result = deployed_contract.functions.singleNestedStruct((1, "string", (True,))).call()
        assert crazy_result == (1, "string", (True,))

    def test_flip_flop(self, w3):
        """Tests single value"""

        deployed_contract = ExampleContract.deploy(
            w3=w3, account=w3.eth.accounts[0], constructor_args=ExampleContract.ConstructorArgs("example")
        )

        flip = 1
        flop = 2
        result = deployed_contract.functions.flipFlop(flip, flop).call()

        assert result.flop == flop
        assert result.flip == flip

    def test_simple_structs(self, w3):
        """Tests single value"""

        deployed_contract = ExampleContract.deploy(
            w3=w3, account=w3.eth.accounts[0], constructor_args=ExampleContract.ConstructorArgs("example")
        )

        input_struct: SimpleStruct = SimpleStruct(1, "string")
        output_struct = deployed_contract.functions.singleSimpleStruct(input_struct).call()

        assert output_struct.intVal == 1
        assert output_struct.strVal == "string"

    def test_nested_structs(self, w3):
        """Tests single value"""

        deployed_contract = ExampleContract.deploy(
            w3=w3, account=w3.eth.accounts[0], constructor_args=ExampleContract.ConstructorArgs("example")
        )

        input_struct = NestedStruct(1, "string", InnerStruct(True))
        output_struct: NestedStruct = deployed_contract.functions.singleNestedStruct(input_struct).call()

        assert output_struct.intVal == 1
        assert output_struct.strVal == "string"
        assert output_struct.innerStruct == InnerStruct(boolVal=True)



"""Dataclasses for all structs in the ContractB contract.

DO NOT EDIT.  This file was generated by pypechain v0.0.46.
See documentation at https://github.com/delvtech/pypechain """

# super() call methods are generic, while our version adds values & types
# pylint: disable=arguments-differ

# contracts have PascalCase names
# pylint: disable=invalid-name

# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments

# unable to determine which imports will be used in the generated code
# pylint: disable=unused-import

# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return

# We import this contract itself to ensure all nested structs have a fully qualified name.
# We use this to avoid namespace collisions, as well as having a uniform
# type structure to do lookups when functions return these structs.
# pylint: disable=import-self


from __future__ import annotations

from dataclasses import dataclass

from ..ContractA import ContractATypes as ContractA
from ..ContractB import ContractBTypes as ContractB


@dataclass
class InnerStruct:
    """InnerStruct struct."""

    intVal: int
    strVal: str


@dataclass
class StructsB:
    """StructsB struct."""

    structA: ContractA.StructsA
    structB: ContractB.InnerStruct



"""Types for foundry-rs."""

from typing import Any, Literal, TypedDict

from eth_typing import ABI


class FoundryByteCode(TypedDict):
    """Foundry"""

    object: str
    sourceMap: str
    linkReferences: dict[str, dict[str, Any]]


class FoundryDeployedByteCode(TypedDict):
    """Foundry"""

    object: str
    sourceMap: str
    linkReferences: Any


class FoundryCompiler(TypedDict):
    """Foundry"""

    version: str


class FoundryMetadata(TypedDict, total=False):
    """Foundry"""

    compiler: FoundryCompiler
    language: Literal["Solidity", "Vyper"]


class FoundryJson(TypedDict):
    """Foundry"""

    abi: ABI
    bytecode: FoundryByteCode
    deployedBytecode: FoundryDeployedByteCode
    methodIdentifiers: dict[str, str]
    rawMetadata: str
    metadata: FoundryMetadata
    ast: Any
    id: int



"""Utilities to help with Solidity types."""

import logging
from typing import TypedDict


class RenderOutput(TypedDict):
    """Defines the file and objects to be exposed."""

    filename: str
    definitions: list[str]


class LinkReferences(TypedDict):
    """Define the structure of the link_references dictionary"""

    contract_name: str
    placeholder_code: str


class SignatureData(TypedDict):
    """Define the structure of the signature_datas dictionary"""

    input_names_and_types: list[str]
    input_names: list[str]
    input_types: list[str]
    outputs: list[str]
    output_types: list[str]


class LinkReferencesData(TypedDict):
    """Define the structure of the link_references_data dictionary"""

    contract_names: list[str]
    contract_names_and_types: list[str]
    contract_types: list[str]
    contract_names_to_placeholder_lookup: list[str]


class FunctionData(TypedDict):
    """Define the structure of the function_data dictionary"""

    name: str
    capitalized_name: str
    signature_datas: list[SignatureData]
    has_overloading: bool
    has_multiple_return_signatures: bool
    has_multiple_return_values: bool


class EventData(TypedDict):
    """Define the structure of the event_data dictionary"""

    name: str
    capitalized_name: str


# pylint: disable=dangerous-default-value
def solidity_to_python_type(solidity_type: str, custom_types: list[str] = []) -> str:
    """Returns the stringified Python type for the given Solidity type.

    Parameters
    ----------
    solidity_type : str
        A solidity variable type string, i.e. 'uint8'...'uint256', 'bool', 'address',
        'bytes2'...'bytes32' etc.

    Returns
    -------
        A python variable type string, i.e. 'int', 'bool', 'address'
    """
    # TODO: use an exhaustive match statement to cover all cases.
    # This isn't worth breaking up into smaller functions, its easy to parse the if statements.
    # pylint: disable=too-many-return-statements
    # pylint: disable=too-many-branches

    # uints and ints

    if solidity_type in [
        "uint8",
        "uint16",
        "uint24",
        "uint32",
        "uint40",
        "uint48",
        "uint56",
        "uint64",
        "uint72",
        "uint80",
        "uint88",
        "uint96",
        "uint104",
        "uint112",
        "uint120",
        "uint128",
        "uint136",
        "uint144",
        "uint152",
        "uint160",
        "uint168",
        "uint176",
        "uint184",
        "uint192",
        "uint200",
        "uint208",
        "uint216",
        "uint232",
        "uint240",
        "uint248",
        "uint256",
        "int8",
        "int16",
        "int24",
        "int32",
        "int40",
        "int48",
        "int56",
        "int64",
        "int72",
        "int80",
        "int88",
        "int96",
        "int104",
        "int112",
        "int120",
        "int128",
        "int136",
        "int144",
        "int152",
        "int160",
        "int168",
        "int176",
        "int184",
        "int192",
        "int200",
        "int208",
        "int216",
        "int224",
        "int232",
        "int240",
        "int248",
        "int256",
    ]:
        return "int"
    # Fixed-size arrays of uints and ints
    if any(solidity_type.startswith(x) for x in ["uint", "int"]) and solidity_type.endswith("]"):
        # TODO: use a package like 'array' or 'numpy' to provide fixed arrays.
        # Extract the size of the array, e.g., "uint8[3]" -> 3
        # size = int(solidity_type.split("[")[-1].split("]")[0])
        # Return a list of 'int' of the given size
        return "list[int]"

    # addresses
    if solidity_type == "address":
        return "str"
    if solidity_type == "address[]":
        return "list[str]"

    # bools
    if solidity_type == "bool":
        return "bool"
    if solidity_type == "bool[]":
        return "list[bool]"

    # bytes
    # for bytes1[], bytes2[],...,bytes32[]
    if solidity_type == "bytes":
        return "bytes"
        # TODO this should actually be a BytesLike string or something.
    if solidity_type.startswith("bytes") and solidity_type.endswith("]"):
        return "list[bytes]"
    # for bytes1, bytes2,...,bytes32
    if solidity_type.startswith("bytes"):
        return "bytes"

    # strings
    if solidity_type == "string":
        return "str"
    if solidity_type == "string[]":
        return "list[str]"
    if solidity_type.startswith("string") and solidity_type.endswith("]"):
        # possible to have e.g. "string[2][]" as the type
        list_depth = solidity_type.count("[")
        out_type_string = ""
        for _ in range(list_depth):
            out_type_string += "list["
        out_type_string += "str" + "]" * list_depth
        return out_type_string

    # tuple
    if solidity_type == "tuple":
        return "tuple"
    if solidity_type == "tuple[]":
        return "list[tuple]"

    # custom types
    if solidity_type in custom_types:
        return solidity_type

    # array of custom types
    if solidity_type[-2:] == "[]" and solidity_type[:-2] in custom_types:
        return "list[" + solidity_type[:-2] + "]"

    # If the Solidity type isn't recognized, make a warning.  This can happen when an internal type
    # is expected for an input parameter or returned in an output.
    logging.warning("Unknown Solidity type: %s", solidity_type)

    return solidity_type


def gather_matching_types(function_datas: list[FunctionData], known_types: list[str]) -> list[str]:
    """Gather matching types from inputs and outputs in the function_datas.

    Parameters
    ----------
    function_datas : list[FunctionData]
        A list of function datas.
    known_types : list[str]
        A list of known types.

    Returns
    -------
    list[str]
        The matching list of types.
    """
    matching_types = []

    for function_data in function_datas:
        for signature_data in function_data["signature_datas"]:
            # Check input types
            for input_type in signature_data["input_types"]:
                if input_type in known_types:
                    matching_types.append(input_type)

            # Check output types
            for output_type in signature_data["output_types"]:
                if output_type in known_types:
                    matching_types.append(output_type)

    return list(set(matching_types))



"""Tests for overloading methods."""

from __future__ import annotations

import os

import pytest
from web3.exceptions import MismatchedABI

from .types.OverloadedMethods import OverloadedMethodsContract, OverloadedMethodsTypes

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


@pytest.mark.usefixtures("process_contracts")
class TestOverloading:
    """Tests pipeline from bots making trades to viewing the trades in the db"""

    def test_overloading(self, w3):
        """Runs the entire pipeline and checks the database at the end.
        All arguments are fixtures.
        """

        deployed_contract = OverloadedMethodsContract.deploy(w3=w3, account=w3.eth.accounts[0])

        s = "test string"
        x = 2
        y = 1

        result = deployed_contract.functions.doSomething().call()
        assert result == deployed_contract.address

        result = deployed_contract.functions.doSomething(s).call()
        assert result == "test string"

        result = deployed_contract.functions.doSomething(x).call()
        assert result == 2 * 2

        result = deployed_contract.functions.doSomething(x, y).call()
        # Expected result is integer division
        assert result == 2 // 1

        # Test kwargs, we pass arguments reversed, but with kwargs
        # so we still expect the same result as above
        result = deployed_contract.functions.doSomething(_y=y, _x=x).call()
        # Expected result is integer division
        assert result == 2 // 1

        result = deployed_contract.functions.doSomething(x, s).call()
        assert isinstance(result, deployed_contract.functions.doSomething(x, s).ReturnValues)
        assert result == deployed_contract.functions.doSomething(x, s).ReturnValues(x, s)

        # Test kwargs, we pass arguments reversed, but with kwargs
        # so we still expect the same result as above
        result = deployed_contract.functions.doSomething(_s=s, _x=x).call()
        assert isinstance(result, deployed_contract.functions.doSomething(x, s).ReturnValues)
        assert result == deployed_contract.functions.doSomething(x, s).ReturnValues(x, s)

        # Test remapped names but different types
        result = deployed_contract.functions.doSomething(s, x).call()
        assert isinstance(result, deployed_contract.functions.doSomething(s, x).ReturnValues)
        assert result == deployed_contract.functions.doSomething(s, x).ReturnValues(s, x)

        # Test kwargs, we pass arguments reversed, but with kwargs
        # so we still expect the same result as above
        result = deployed_contract.functions.doSomething(_s=x, _x=s).call()
        assert isinstance(result, deployed_contract.functions.doSomething(s, x).ReturnValues)
        assert result == deployed_contract.functions.doSomething(s, x).ReturnValues(s, x)

        result = deployed_contract.functions.doSomething(OverloadedMethodsTypes.SimpleStruct(s, x)).call()
        assert isinstance(result, OverloadedMethodsTypes.SimpleStruct)
        assert result == OverloadedMethodsTypes.SimpleStruct(s, x)

        # Test kwargs with structs
        result = deployed_contract.functions.doSomething(_simpleStruct=OverloadedMethodsTypes.SimpleStruct(s, x)).call()
        assert isinstance(result, OverloadedMethodsTypes.SimpleStruct)
        assert result == OverloadedMethodsTypes.SimpleStruct(s, x)

        result = deployed_contract.functions.doSomething(
            [
                OverloadedMethodsTypes.SimpleStruct(s, x),
                OverloadedMethodsTypes.SimpleStruct(s, y),
            ]
        ).call()
        # TODO although the typing of this function says it returns a list,
        # it actually returns a tuple. Fix the output type of this function.
        # assert isinstance(result, list)
        for r in result:
            assert isinstance(r, OverloadedMethodsTypes.SimpleStruct)
        assert result[0] == OverloadedMethodsTypes.SimpleStruct(s, x)
        assert result[1] == OverloadedMethodsTypes.SimpleStruct(s, y)

        result = deployed_contract.functions.doSomething(
            nestedStructVec=[
                OverloadedMethodsTypes.NestedStruct(x, s, OverloadedMethodsTypes.SimpleStruct(s, x)),
                OverloadedMethodsTypes.NestedStruct(y, s, OverloadedMethodsTypes.SimpleStruct(s, y)),
            ],
        ).call()
        # TODO although the typing of this function says it returns a list,
        # it actually returns a tuple. Fix the output type of this function.
        # assert isinstance(result, list)
        for r in result:
            assert isinstance(r, OverloadedMethodsTypes.NestedStruct)
        assert result[0] == OverloadedMethodsTypes.NestedStruct(x, s, OverloadedMethodsTypes.SimpleStruct(s, x))
        assert result[1] == OverloadedMethodsTypes.NestedStruct(y, s, OverloadedMethodsTypes.SimpleStruct(s, y))

        with pytest.raises(MismatchedABI) as err:
            result = deployed_contract.functions.doSomething(x, y, s).call()  # type: ignore
        assert "Could not identify the intended function with name `doSomething`" in str(err.value)

    def test_factory(self, w3):
        """Tests creating multiple contracts with the same object"""
        # Deploy two contracts
        deployed_contract_1 = OverloadedMethodsContract.deploy(w3=w3, account=w3.eth.accounts[0])
        deployed_contract_2 = OverloadedMethodsContract.deploy(w3=w3, account=w3.eth.accounts[0])

        # Function overload with no arguments returns the address
        # Ensure the referenced contract is actually the contract being called
        result = deployed_contract_1.functions.doSomething().call()
        assert result == deployed_contract_1.address
        result = deployed_contract_2.functions.doSomething().call()
        assert result == deployed_contract_2.address



"""Tests for rendering the contract file."""

import json
import os

from eth_typing import ABI

from pypechain.render.contract import get_function_datas
from pypechain.utilities.abi import get_events_for_abi
from pypechain.utilities.templates import get_jinja_env

# using pytest fixtures necessitates this.
# pylint: disable=redefined-outer-name

current_path = os.path.abspath(os.path.dirname(__file__))
project_root = os.path.dirname(os.path.dirname(current_path))


class TestOverloading:
    """Tests overloading for code gen using snapshots"""

    def test_overloading(self, snapshot):
        """Tests overloading for code gen using snapshots"""

        env = get_jinja_env()
        functions_template = env.get_template("contract.py/functions.py.jinja2")

        # TODO: add return types to function calls

        abi_str = """
            [
                {
                    "constant": true,
                    "inputs": [],
                    "name": "balanceOf",
                    "outputs": [{"name": "", "type": "uint256"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                },
                {
                    "constant": true,
                    "inputs": [{"name": "who", "type": "address"}],
                    "name": "balanceOf",
                    "outputs": [{"name": "", "type": "uint256"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                }
            ]
        """

        abi: ABI = json.loads(abi_str)

        function_datas, constructor_data = get_function_datas(abi)
        has_overloading = any(len(function_data["signature_datas"]) > 1 for function_data in function_datas.values())
        contract_name = "Overloaded"

        functions_block = functions_template.render(
            abi=abi,
            contract_name=contract_name,
            functions=function_datas,
            # TODO: use this data to add a typed constructor
            constructor=constructor_data,
        )
        assert has_overloading is True

        snapshot.snapshot_dir = "snapshots"  # This line is optional.
        snapshot.assert_match(functions_block, "expected_overloading.py")

    def test_notoverloading(self, snapshot):
        """Tests not overloading for code gen using snapshots"""

        env = get_jinja_env()
        functions_template = env.get_template("contract.py/functions.py.jinja2")

        # TODO: add return types to function calls

        # different names, should NOT be overloaded
        abi_str = """
            [
                {
                    "constant": true,
                    "inputs": [],
                    "name": "balanceOf",
                    "outputs": [{"name": "", "type": "uint256"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                },
                {
                    "constant": true,
                    "inputs": [{"name": "who", "type": "address"}],
                    "name": "balanceOfWho",
                    "outputs": [{"name": "", "type": "bool"}],
                    "payable": false,
                    "stateMutability": "view",
                    "type": "function"
                }
            ]
        """

        abi: ABI = json.loads(abi_str)

        function_datas, constructor_data = get_function_datas(abi)
        has_overloading = any(len(function_data["signature_datas"]) > 1 for function_data in function_datas.values())
        contract_name = "Overloaded"

        functions_block = functions_template.render(
            abi=abi,
            contract_name=contract_name,
            functions=function_datas,
            # TODO: use this data to add a typed constructor
            constructor=constructor_data,
        )
        assert has_overloading is False

        snapshot.snapshot_dir = "snapshots"
        snapshot.assert_match(functions_block, "expected_not_overloading.py")

    def test_has_events(self, snapshot):
        """Tests that events are created."""

        env = get_jinja_env()
        events_template = env.get_template("contract.py/events.py.jinja2")

        # different names, should NOT be overloaded
        abi_str = """
            [
                {
                    "anonymous": false,
                    "inputs": [
                        {
                        "indexed": true,
                        "internalType": "address",
                        "name": "from",
                        "type": "address"
                        },
                        {
                        "indexed": true,
                        "internalType": "address",
                        "name": "to",
                        "type": "address"
                        },
                        {
                        "indexed": false,
                        "internalType": "uint256",
                        "name": "value",
                        "type": "uint256"
                        }
                    ],
                    "name": "Transfer",
                    "type": "event"
                },
                {
                    "anonymous": false,
                    "inputs": [
                        {
                        "indexed": true,
                        "internalType": "address",
                        "name": "owner",
                        "type": "address"
                        },
                        {
                        "indexed": true,
                        "internalType": "address",
                        "name": "spender",
                        "type": "address"
                        },
                        {
                        "indexed": false,
                        "internalType": "uint256",
                        "name": "value",
                        "type": "uint256"
                        }
                    ],
                    "name": "Approval",
                    "type": "event"
                }
            ]
        """

        abi: ABI = json.loads(abi_str)

        event_datas = get_events_for_abi(abi)
        contract_name = "Overloaded"

        events_block = events_template.render(
            abi=abi,
            contract_name=contract_name,
            events=event_datas,
        )

        snapshot.snapshot_dir = "snapshots"
        snapshot.assert_match(events_block, "has_events.py")



"""Export all types from generated files.

DO NOT EDIT.  This file was generated by pypechain v0.0.46.
See documentation at https://github.com/delvtech/pypechain """

# The module name reflects that of the solidity contract,
# which may not adhere to python naming conventions
# pylint: disable=invalid-name


from .Contract import ContractContract
from .MyLibrary import MyLibraryContract



"""A web3.py Contract class for the Transact contract.

DO NOT EDIT.  This file was generated by pypechain v0.0.46.
See documentation at https://github.com/delvtech/pypechain """

# contracts have PascalCase names
# pylint: disable=invalid-name

# contracts control how many attributes and arguments we have in generated code
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments

# we don't need else statement if the other conditionals all have return,
# but it's easier to generate
# pylint: disable=no-else-return

# This file is bound to get very long depending on contract sizes.
# pylint: disable=too-many-lines

# methods are overridden with specific arguments instead of generic *args, **kwargs
# pylint: disable=arguments-differ

# consumers have too many opinions on line length
# pylint: disable=line-too-long

# We use protected classes within pypechain
# pylint: disable=protected-access

# We sometimes define a variable that might not be returned in `call`,
# but we still may want to call the function
# pylint: disable=unused-variable


from __future__ import annotations

import copy
from typing import Any, Iterable, Sequence, Type, cast, overload

from eth_account.signers.local import LocalAccount
from eth_typing import ABI, ChecksumAddress, HexStr
from hexbytes import HexBytes
from typing_extensions import Self
from web3 import Web3
from web3._utils.events import EventLogErrorFlags
from web3._utils.filters import LogFilter
from web3.contract.contract import Contract, ContractConstructor, ContractEvent, ContractEvents, ContractFunctions
from web3.logs import WARN
from web3.types import BlockIdentifier, StateOverride, TxParams, TxReceipt

from pypechain.core import (
    BaseEventArgs,
    PypechainBaseContractErrors,
    PypechainContractFunction,
    check_txn_receipt,
    combomethod_typed,
    dataclass_to_tuple,
    expand_struct_type_str,
    get_arg_type_names,
    handle_contract_logic_error,
)

from .TransactTypes import SuccessEvent

structs = {}


class TransactTransactFunctionContractFunction0(PypechainContractFunction):
    """ContractFunction for the TransactFunction() method."""

    _function_name = "TransactFunction"
    _type_signature = expand_struct_type_str(tuple([]), structs)

    def call(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier | None = None,
        state_override: StateOverride | None = None,
        ccip_read_enabled: bool | None = None,
    ) -> None:
        """returns None."""
        # We handle the block identifier = None case here for typing.
        if block_identifier is None:
            block_identifier = self.w3.eth.default_block

        # Call the function
        try:
            raw_values = super().call(transaction, block_identifier, state_override, ccip_read_enabled)
        except Exception as err:  # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=TransactContractErrors,
                err=err,
                contract_call_type="call",
                transaction=transaction,
                block_identifier=block_identifier,
            ) from err

    def transact(self, transaction: TxParams | None = None) -> HexBytes:
        try:
            return super().transact(transaction)
        except Exception as err:  # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=TransactContractErrors,
                err=err,
                contract_call_type="transact",
                transaction=transaction,
                block_identifier="pending",  # race condition here, best effort to get block of txn.
            ) from err

    def estimate_gas(
        self,
        transaction: TxParams | None = None,
        block_identifier: BlockIdentifier | None = None,
        state_override: StateOverride | None = None,
    ) -> int:
        try:
            return super().estimate_gas(transaction, block_identifier, state_override)
        except Exception as err:  # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=TransactContractErrors,
                err=err,
                contract_call_type="build",
                transaction=transaction,
                block_identifier="pending",  # race condition here, best effort to get block of txn.
            ) from err

    def build_transaction(self, transaction: TxParams | None = None) -> TxParams:
        try:
            return super().build_transaction(transaction)
        except Exception as err:  # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=TransactContractErrors,
                err=err,
                contract_call_type="build",
                transaction=transaction,
                block_identifier="pending",  # race condition here, best effort to get block of txn.
            ) from err

    def sign_and_transact(self, account: LocalAccount, transaction: TxParams | None = None) -> HexBytes:
        """Convenience method for signing and sending a transaction using the provided account.

        Arguments
        ---------
        account : LocalAccount
            The account to use for signing and sending the transaction.
        transaction : TxParams | None, optional
            The transaction parameters to use for sending the transaction.

        Returns
        -------
        HexBytes
            The transaction hash.
        """
        if transaction is None:
            transaction_params: TxParams = {}
        else:
            transaction_params: TxParams = transaction

        if "from" in transaction_params:
            # Ensure if transaction is set, it matches
            assert (
                transaction_params["from"] == account.address
            ), f"Transaction from {transaction_params['from']} does not match account {account.address}"
        else:
            transaction_params["from"] = account.address

        if "gas" not in transaction_params:
            # Web3 default gas estimate seems to be underestimating gas, likely due to
            # not looking at pending block. Here, we explicitly call estimate gas
            # if gas isn't passed in.
            transaction_params["gas"] = self.estimate_gas(transaction_params, block_identifier="pending")

        # Build the raw transaction
        raw_transaction = self.build_transaction(transaction_params)

        if "nonce" not in raw_transaction:
            raw_transaction["nonce"] = self.w3.eth.get_transaction_count(account.address, block_identifier="pending")

        # Sign the raw transaction
        # Mismatched types between account and web3py
        signed_transaction = account.sign_transaction(raw_transaction)  # type: ignore

        # Send the signed transaction
        try:
            return self.w3.eth.send_raw_transaction(signed_transaction.raw_transaction)
        except Exception as err:  # pylint disable=broad-except
            raise handle_contract_logic_error(
                contract_function=self,
                errors_class=TransactContractErrors,
                err=err,
                contract_call_type="transact",
                transaction=transaction_params,
                block_identifier="pending",  # race condition here, best effort to get block of txn.
            ) from err

    def sign_transact_and_wait(
        self,
        account: LocalAccount,
        transaction: TxParams | None = None,
        timeout: float | None = None,
        poll_latency: float | None = None,
        validate_transaction: bool = False,
    ) -> TxReceipt:
        """Convenience method for signing and sending a transaction using the provided account.

        Arguments
        ---------
        account : LocalAccount
            The account to use for signing and sending the transaction.
        transaction : TxParams | None, optional
            The transaction parameters to use for sending the transaction.
        timeout: float, optional
            The number of seconds to wait for the transaction to be mined. Defaults to 120.
        poll_latency: float, optional
            The number of seconds to wait between polling for the transaction receipt. Defaults to 0.1.
        validate_transaction: bool, optional
            Whether to validate the transaction. If True, will throw an exception if the resulting
            tx_receipt returned a failure status.

        Returns
        -------
        HexBytes
            The transaction hash.
        """

        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-positional-arguments

        if timeout is None:
            timeout = 120
        if poll_latency is None:
            poll_latency = 0.1

        tx_hash = self.sign_and_transact(account, transaction)
        tx_receipt = self.w3.eth.wait_for_transaction_receipt(tx_hash, timeout=timeout, poll_latency=poll_latency)
        # Check the receipt, throwing an error if status == 0
        if validate_transaction:
            return check_txn_receipt(self, tx_hash, tx_receipt)
        else:
            return tx_receipt


class TransactTransactFunctionContractFunction(PypechainContractFunction):
    """ContractFunction for the TransactFunction method."""

    # super() call methods are generic, while our version adds values & types
    # pylint: disable=arguments-differ# disable this warning when there is overloading
    # pylint: disable=function-redefined

    _function_name = "TransactFunction"

    # Make lookup for function signature -> overloaded function
    # The function signatures are python types, as we need to do a
    # lookup of arguments passed in to contract function
    _functions: dict[str, PypechainContractFunction]

    @overload
    def __call__(self) -> TransactTransactFunctionContractFunction0:  # type: ignore
        ...

    def __call__(self, *args, **kwargs) -> TransactTransactFunctionContractFunction:  # type: ignore
        clone = super().__call__(
            *(dataclass_to_tuple(arg) for arg in args), **{key: dataclass_to_tuple(arg) for key, arg in kwargs.items()}
        )

        # Arguments is the flattened set of arguments from args and kwargs, ordered by the abi
        # We get the python types of the args passed in, but remapped from tuples -> dataclasses
        arg_types = get_arg_type_names(clone.arguments)

        # Look up the function class based on arg types.
        # We ensure we use a copy of the original object.
        function_obj = copy.copy(self._functions[arg_types])

        function_obj.args = clone.args
        function_obj.kwargs = clone.kwargs

        # The `@overload` of `__call__` takes care of setting the type of this object correctly
        return function_obj  # type: ignore

    @classmethod
    def factory(cls, class_name: str, **kwargs: Any) -> Self:
        out = super().factory(class_name, **kwargs)

        # We initialize our overridden functions here.
        # Note that we use the initialized object to ensure each function
        # is attached to the instanciated object
        # (attached to a specific web3 and contract address)
        out._functions = {
            TransactTransactFunctionContractFunction0._type_signature: TransactTransactFunctionContractFunction0.factory(
                "TransactTransactFunctionContractFunction0", **kwargs
            ),
        }
        return out


class TransactContractFunctions(ContractFunctions):
    """ContractFunctions for the Transact contract."""

    TransactFunction: TransactTransactFunctionContractFunction

    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: ChecksumAddress | None = None,
        decode_tuples: bool | None = False,
    ) -> None:
        super().__init__(abi, w3, address, decode_tuples)
        self.TransactFunction = TransactTransactFunctionContractFunction.factory(
            "TransactFunction",
            w3=w3,
            contract_abi=abi,
            address=address,
            decode_tuples=decode_tuples,
            abi_element_identifier="TransactFunction",
        )


class TransactSuccessContractEvent(ContractEvent):
    """ContractEvent for Success."""

    # super() get_logs and create_filter methods are generic, while our version adds values & types
    # pylint: disable=arguments-differ

    # pylint: disable=useless-parent-delegation
    def __init__(self, *argument_names: tuple[str]) -> None:
        super().__init__(*argument_names)

    @combomethod_typed
    def get_logs_typed(
        self,
        argument_filters: dict[str, Any] | None = None,
        from_block: BlockIdentifier | None = None,
        to_block: BlockIdentifier | None = None,
        block_hash: HexBytes | None = None,
    ) -> Iterable[SuccessEvent]:
        """Extension of `get_logs` that return a typed dataclass of the event."""
        abi_events = super().get_logs(
            argument_filters=argument_filters, from_block=from_block, to_block=to_block, block_hash=block_hash
        )
        # TODO there may be issues with this function if the user uses a middleware that changes event structure.
        return [
            SuccessEvent(
                log_index=abi_event.logIndex,
                transaction_index=abi_event.transactionIndex,
                transaction_hash=abi_event.transactionHash,
                address=abi_event.address,
                block_hash=abi_event.blockHash,
                block_number=abi_event.blockNumber,
                args=BaseEventArgs(),
            )
            for abi_event in abi_events
        ]

    @combomethod_typed
    def process_receipt_typed(
        self, txn_receipt: TxReceipt, errors: EventLogErrorFlags = WARN
    ) -> Iterable[SuccessEvent]:
        """Extension of `process_receipt` that return a typed dataclass of the event."""
        abi_events = super().process_receipt(txn_receipt, errors)
        # TODO there may be issues with this function if the user uses a middleware that changes event structure.
        return [
            SuccessEvent(
                log_index=abi_event.logIndex,
                transaction_index=abi_event.transactionIndex,
                transaction_hash=abi_event.transactionHash,
                address=abi_event.address,
                block_hash=abi_event.blockHash,
                block_number=abi_event.blockNumber,
                args=BaseEventArgs(),
            )
            for abi_event in abi_events
        ]

    @combomethod_typed
    def create_filter(  # type: ignore
        self,
        *,  # PEP 3102
        argument_filters: dict[str, Any] | None = None,
        from_block: BlockIdentifier | None = None,
        to_block: BlockIdentifier = "latest",
        address: ChecksumAddress | None = None,
        topics: Sequence[Any] | None = None,
    ) -> LogFilter:
        return cast(
            LogFilter,
            super().create_filter(
                argument_filters=argument_filters,
                from_block=from_block,
                to_block=to_block,
                address=address,
                topics=topics,
            ),
        )


class TransactContractEvents(ContractEvents):
    """ContractEvents for the Transact contract."""

    Success: Type[TransactSuccessContractEvent]

    def __init__(
        self,
        abi: ABI,
        w3: "Web3",
        address: ChecksumAddress | None = None,
    ) -> None:
        super().__init__(abi, w3, address)
        self.Success = cast(
            Type[TransactSuccessContractEvent],
            TransactSuccessContractEvent.factory(
                "Success", w3=w3, contract_abi=abi, address=address, event_name="Success"
            ),
        )


transact_abi: ABI = cast(
    ABI,
    [
        {"type": "function", "name": "TransactFunction", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
        {"type": "event", "name": "Success", "inputs": [], "anonymous": False},
    ],
)


class TransactContractErrors(PypechainBaseContractErrors):
    """ContractErrors for the Transact contract."""

    def __init__(
        self,
    ) -> None:

        self._all = []


class TransactContract(Contract):
    """A web3.py Contract class for the Transact contract."""

    abi: ABI = transact_abi
    # We change `bytecode` as needed for linking, but keep
    # `_raw_bytecode` unchanged as an original copy.
    # pylint: disable=line-too-long
    _raw_bytecode: HexStr | None = HexStr(
        "0x6080604052348015600f57600080fd5b5060968061001e6000396000f3fe6080604052348015600f57600080fd5b506004361060285760003560e01c8063e437702614602d575b600080fd5b60336035565b005b6040517f395a9ab3d1230297d931e1fa224ca597ca0e45f620c1aeb74b512bfcc6f66aab90600090a156fea26469706673582212205efd683901ebe890a4751a30eb78d980861b5e0a81ad24a35489c4a1eeb9fe3764736f6c63430008160033"
    )

    def __init__(self, address: ChecksumAddress | None = None) -> None:
        # Initialize parent Contract class
        super().__init__(address=address)
        self.functions = TransactContractFunctions(transact_abi, self.w3, address)  # type: ignore
        self.events = TransactContractEvents(transact_abi, self.w3, address)  # type: ignore

    events: TransactContractEvents

    functions: TransactContractFunctions

    @classmethod
    def constructor(cls) -> ContractConstructor:  # type: ignore
        """Creates a transaction with the contract's constructor function.

        Parameters
        ----------

        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.

        """
        cls.bytecode = cls._raw_bytecode
        if cls.bytecode is not None:

            # bytecode needs to be in hex for web3
            cls.bytecode = HexBytes(cls.bytecode)

        return super().constructor()

    @classmethod
    def deploy(cls, w3: Web3, account: LocalAccount | ChecksumAddress) -> Self:
        """Deploys an instance of the contract.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        account : LocalAccount
            The account to use to deploy the contract.

        Returns
        -------
        Self
            A deployed instance of the contract.
        """
        deployer = cls.factory(w3=w3)
        constructor_fn = deployer.constructor()

        # if an address is supplied, try to use a web3 default account
        if isinstance(account, str):
            tx_hash = constructor_fn.transact({"from": account})
            tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

            deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
            return deployed_contract

        # otherwise use the account provided.
        deployment_tx = constructor_fn.build_transaction()
        current_nonce = w3.eth.get_transaction_count(account.address, "pending")
        deployment_tx.update({"nonce": current_nonce})

        # Sign the transaction with local account private key
        signed_tx = account.sign_transaction(deployment_tx)

        # Send the signed transaction and wait for receipt
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)

        deployed_contract = deployer(address=tx_receipt.contractAddress)  # type: ignore
        return deployed_contract

    @classmethod
    def factory(cls, w3: Web3, class_name: str | None = None, **kwargs: Any) -> Type[Self]:
        """Initializes the contract object.

        Parameters
        ----------
        w3 : Web3
            A web3 instance.
        class_name: str | None
            The instance class name.

        Returns
        -------
        Self
            An instance of the contract class.
        """
        contract = super().factory(w3, class_name, **kwargs)
        contract.functions = TransactContractFunctions(transact_abi, w3, None)

        return contract
